# -*- coding: utf-8 -*-
# file: generate_classification_dataset.py
# time: 2021/7/31
# author: yangheng <yangheng@m.scnu.edu.cn>
# github: https://github.com/yangheng95
# Copyright (C) 2021. All Rights Reserved.
import os

import tqdm
from gensim import corpora, models, similarities
import os.path

import global_config
from injection.injection_generation import generate_injection_cases


class SimilarityCalculator:
    def __init__(self, task, num_topics=100):
        self.corpus = None
        self.dic = None
        self.contents = []

        self.contents = load_injections(task)
        self.prepare_dic_corpus()

        self.lsi = models.LsiModel(self.corpus, id2word=self.dic, num_topics=num_topics)
        # self.index = similarities.Similarity('index', self.lsi[self.corpus], num_features=self.lsi.num_topics)  c````
        self.index = similarities.MatrixSimilarity(self.lsi[self.corpus])
        self.matched = set()

    def prepare_dic_corpus(self):
        texts = self.contents
        dic = corpora.Dictionary([content[1].split() for content in self.contents])
        corpus = [dic.doc2bow(text[1].split()) for text in texts]
        self.corpus = corpus
        self.dic = dic
        return dic, corpus

    def find_similar_pairs(self, query_injection):
        query_bow = self.dic.doc2bow(query_injection.lower().split())
        query_lsi = self.lsi[query_bow]
        sims = self.index[query_lsi]
        sorted_sims = sorted(enumerate(sims), key=lambda item: -item[1])

        # print('max similarity:{}'.format(sorted_sims[0][1]))
        # print('min similarity:{}'.format(sorted_sims[-1][1]))
        if sorted_sims[0][1] > 0:
            while len(sorted_sims) > 1 and self.contents[sorted_sims[0][0]][1] in self.matched:
                sorted_sims.remove(sorted_sims[0])
            self.matched.add(self.contents[sorted_sims[0][0]][1])
            return query_injection, self.contents[sorted_sims[0][0]][1]
        else:
            return '', ''


def extract_translator_dataset(source_task, target_task):
    query_injections = load_injections(target_task)

    sim_cal = SimilarityCalculator(source_task)

    dir_name = os.path.join(global_config.PROJECT_PATH, 'deep_learning/translator/translation_datasets')
    dir_to_save = os.path.join(dir_name, '{}2{}'.format(source_task, target_task))

    if not os.path.exists(dir_to_save):
        os.makedirs(dir_to_save)

    source_lan_path = os.path.join(dir_name, '{}/{}2{}_source.txt'.format(dir_to_save, source_task, target_task))
    target_lan_path = os.path.join(dir_name, '{}/{}2{}_target.txt'.format(dir_to_save, source_task, target_task))

    f_src_out = open(source_lan_path, mode='w', encoding='utf8')
    f_tgt_out = open(target_lan_path, mode='w', encoding='utf8')
    source_injection_set = set()
    target_injection_set = set()
    for _, query in tqdm.tqdm(query_injections, postfix='extract similar injection {} to {}'.format(source_task, target_task)):
        source_injection, target_injection = sim_cal.find_similar_pairs(query)
        if source_injection and target_injection:
            source_injection_set.add(source_injection)
            target_injection_set.add(target_injection)

            f_src_out.write(source_injection + '\n')
            f_tgt_out.write(target_injection + '\n')

        if len(source_injection_set) == len(query_injections):
            break
    print('source_injections{}'.format(len(source_injection_set)))
    print('target_injections{}'.format(len(target_injection_set)))


def load_injections(task, itype='pass'):
    fname = '{}_pass.txt'.format(task) if itype == 'pass' else '{}.txt'.format(task)
    data_dir = os.path.join(os.getcwd(), 'injection_cases/{}/{}'.format(task, fname))
    fin = open(data_dir, mode='r', encoding='utf8')

    contents = []
    for i, line in enumerate(fin.readlines()):
        contents.append((i, line.split('$BYPASS_LABEL$')[0].strip()))
    fin.close()
    return contents


if __name__ == '__main__':
    tasks = ["sqli", "xss", "osi", "phpi", "xmli", "htmli"]

    for task1 in tasks:
        for task2 in tasks:
            if task1 != task2:
                extract_translator_dataset(task1, task2)

    # for debug
    task1, task2 = "xss", "osi"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "osi", "phpi"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "phpi", "xmli"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "xmli", "htmli"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "sqli", "osi"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "xss", "phpi"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "osi", "xmli"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "phpi", "htmli"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "xss", "xmli"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "osi", "htmli"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "sqli", "xmli",
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "xss", "htmli"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "sqli", "htmli"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "sqli", "phpi"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
    task1, task2 = "sqli", "xss"
    extract_translator_dataset(task1, task2)
    extract_translator_dataset(task2, task1)
